@@ -32,6 +32,8 @@ char tls_library_version[] = "Schannel";
 #define PROT_TLS1_2 4
 #define PROT_TLS1_3 8
 
+static int ma_check_peer_cert_time(MARIADB_TLS *ctls);
+
 static struct
 {
   DWORD cipher_id;
@@ -374,7 +376,6 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
   ALG_ID AlgId[MAX_ALG_ID];
   size_t i;
   DWORD protocol = 0;
-  int verify_certs;
   const CERT_CONTEXT* cert_context = NULL;
 
   if (!ctls)
@@ -421,15 +422,13 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
 
   if (mysql->options.extension && mysql->options.extension->tls_version)
   {
-    if (strstr(mysql->options.extension->tls_version, "TLSv1.0"))
-      Cred.grbitEnabledProtocols|= SP_PROT_TLS1_0_CLIENT;
     if (strstr(mysql->options.extension->tls_version, "TLSv1.1"))
       Cred.grbitEnabledProtocols|= SP_PROT_TLS1_1_CLIENT;
     if (strstr(mysql->options.extension->tls_version, "TLSv1.2"))
       Cred.grbitEnabledProtocols|= SP_PROT_TLS1_2_CLIENT;
   }
   if (!Cred.grbitEnabledProtocols)
-    Cred.grbitEnabledProtocols = SP_PROT_TLS1_0_CLIENT | SP_PROT_TLS1_1_CLIENT | SP_PROT_TLS1_2_CLIENT;
+    Cred.grbitEnabledProtocols = SP_PROT_TLS1_1_CLIENT | SP_PROT_TLS1_2_CLIENT;
 
 
   if (ma_tls_set_client_certs(ctls, &cert_context))
@@ -450,15 +449,6 @@ my_bool ma_tls_connect(MARIADB_TLS *ctls)
   if (ma_schannel_client_handshake(ctls) != SEC_E_OK)
     goto end;
 
-   verify_certs =  mysql->options.ssl_ca || mysql->options.ssl_capath ||
-     !mysql->options.extension->tls_allow_invalid_server_cert;
-
-  if (verify_certs)
-  {
-    if (!ma_schannel_verify_certs(ctls, !mysql->options.extension->tls_allow_invalid_server_cert))
-      goto end;
-  }
-
   rc = 0;
 
 end:
@@ -513,14 +503,40 @@ my_bool ma_tls_close(MARIADB_TLS *ctls)
       DeleteSecurityContext(&sctx->hCtxt);
   }
   LocalFree(sctx);
+  LocalFree(ctls->cert_info.issuer);
+  LocalFree(ctls->cert_info.subject);
   return 0;
 }
 /* }}} */
 
-int ma_tls_verify_server_cert(MARIADB_TLS *ctls)
+int ma_tls_verify_server_cert(MARIADB_TLS *ctls, unsigned int verify_flags)
 {
-  /* Done elsewhere */
-  return 0;
+  MYSQL *mysql;
+  SC_CTX *sctx;
+  if (!ctls || !ctls->ssl || !ctls->pvio || !ctls->pvio->mysql)
+    return 1;
+
+  sctx= (SC_CTX *)ctls->ssl; 
+  mysql= ctls->pvio->mysql;
+
+  if (verify_flags & MARIADB_TLS_VERIFY_PERIOD)
+  {
+    if (ma_check_peer_cert_time(ctls))
+    {
+      mysql->net.tls_verify_status|= MARIADB_TLS_VERIFY_PERIOD;
+      return 1;
+    }
+  }
+
+  if (verify_flags & MARIADB_TLS_VERIFY_FINGERPRINT)
+  {
+    if (ma_pvio_tls_check_fp(ctls, mysql->options.extension->tls_fp, mysql->options.extension->tls_fp_list))
+    {
+      mysql->net.tls_verify_status |= MARIADB_TLS_VERIFY_FINGERPRINT;
+      return 1;
+    }
+  }
+  return ma_schannel_verify_certs(ctls, verify_flags);
 }
 
 static const char *cipher_name(const SecPkgContext_CipherInfo *CipherInfo)
@@ -553,10 +569,118 @@ const char *ma_tls_get_cipher(MARIADB_TLS *ctls)
   return cipher_name(&CipherInfo);
 }
 
+unsigned char *ma_cert_blob_to_str(PCERT_NAME_BLOB cnblob)
+{
+  DWORD type= CERT_X500_NAME_STR;
+  DWORD size= CertNameToStrA(X509_ASN_ENCODING, cnblob, type, NULL, 0);
+  char *str= NULL;
+ 
+  if (!size)
+    return NULL;
+
+  str= (char *)LocalAlloc(LMEM_ZEROINIT,size);
+  CertNameToStrA(X509_ASN_ENCODING, cnblob, type, str, size);
+  return str;
+}
+
+static int ma_check_peer_cert_time(MARIADB_TLS *ctls)
+{
+  PCCERT_CONTEXT pCertCtx= NULL;
+  SC_CTX *sctx;
+  PCERT_INFO pci= NULL;
+  FILETIME ft;
+  SYSTEMTIME st;
+
+  if (!ctls || !ctls->ssl || !ctls->pvio || !ctls->pvio->mysql)
+    return 1;
+  
+  sctx= (SC_CTX *)ctls->ssl;
+
+  if (QueryContextAttributes(&sctx->hCtxt, SECPKG_ATTR_REMOTE_CERT_CONTEXT, (PVOID)&pCertCtx) != SEC_E_OK)
+    return 1;
+
+  pci= pCertCtx->pCertInfo;
+
+  GetSystemTime(&st);
+  SystemTimeToFileTime(&st, &ft);
+
+  if (CompareFileTime(&ft, &pci->NotBefore) == -1 ||
+      CompareFileTime(&pci->NotAfter, &ft) == -1)
+    return 1;
+
+  return 0;
+}
+
+static void ma_systime_to_tm(SYSTEMTIME sys_tm, struct tm *tm)
+{
+  memset(tm, 0, sizeof(struct tm));
+  tm->tm_year= sys_tm.wYear - 1900;
+  tm->tm_mon= sys_tm.wMonth - 1;
+  tm->tm_mday= sys_tm.wDay;
+  tm->tm_hour = sys_tm.wHour;
+  tm->tm_min = sys_tm.wMinute;
+}
+
+unsigned int ma_tls_get_peer_cert_info(MARIADB_TLS *ctls, unsigned int hash_size)
+{
+  PCCERT_CONTEXT pCertCtx= NULL;
+  SC_CTX *sctx;
+  PCERT_INFO pci= NULL;
+  DWORD size´= 0;
+  SYSTEMTIME tm;
+  char fp[129];
+  unsigned int hash_alg;
+
+  if (!ctls || !ctls->ssl || !ctls->pvio || !ctls->pvio->mysql)
+    return 1;
+
+  sctx= (SC_CTX *)ctls->ssl;
+
+  switch (hash_size) {
+    case 0:
+    case 256:
+      hash_alg= MA_HASH_SHA256;
+      break;
+    case 384:
+      hash_alg= MA_HASH_SHA384;
+      break;
+    case 512:
+      hash_alg= MA_HASH_SHA512;
+      break;
+    default:
+      my_set_error(ctls->pvio->mysql, CR_SSL_CONNECTION_ERROR, SQLSTATE_UNKNOWN,
+                   ER(CR_SSL_CONNECTION_ERROR),
+                   "Cannot detect hash algorithm for fingerprint verification");
+      return 1;
+  }
+
+  /* Did we already read peer cert information ? */
+  if (!ctls->cert_info.version)
+  {
+    if (QueryContextAttributes(&sctx->hCtxt, SECPKG_ATTR_REMOTE_CERT_CONTEXT, (PVOID)&pCertCtx) != SEC_E_OK)
+      return 1;
+
+    pci= pCertCtx->pCertInfo;
+
+    ctls->cert_info.version= pci->dwVersion + 1;
+    ctls->cert_info.subject = ma_cert_blob_to_str(&pci->Subject);
+    ctls->cert_info.issuer = ma_cert_blob_to_str(&pci->Issuer);
+
+    FileTimeToSystemTime(&pci->NotBefore, &tm);
+    ma_systime_to_tm(tm, &ctls->cert_info.not_before);
+    FileTimeToSystemTime(&pci->NotAfter, &tm);
+    ma_systime_to_tm(tm, &ctls->cert_info.not_after);
+  }
+  ma_tls_get_finger_print(ctls, hash_alg, fp, sizeof(fp));
+  mysql_hex_string(ctls->cert_info.fingerprint, fp, (unsigned long)ma_hash_digest_size(hash_alg));
+
+  return 0; 
+}
+
+
 unsigned int ma_tls_get_finger_print(MARIADB_TLS *ctls, uint hash_type, char *fp, unsigned int len)
 {
   MA_HASH_CTX* hash_ctx;
-  uint hash_len;
 
   SC_CTX *sctx= (SC_CTX *)ctls->ssl;
   PCCERT_CONTEXT pRemoteCertContext = NULL;
@@ -574,14 +698,13 @@ unsigned int ma_tls_get_finger_print(MARIADB_TLS *ctls, uint hash_type, char *fp
   if (QueryContextAttributes(&sctx->hCtxt, SECPKG_ATTR_REMOTE_CERT_CONTEXT, (PVOID)&pRemoteCertContext) != SEC_E_OK)
     return 0;
 
-  hash_ctx = ma_hash_new(hash_type, NULL);
+  hash_ctx = ma_hash_new(hash_type);
   ma_hash_input(hash_ctx, pRemoteCertContext->pbCertEncoded, pRemoteCertContext->cbCertEncoded);
   ma_hash_result(hash_ctx, fp);
-  hash_len = hash_ctx->digest_len;
   ma_hash_free(hash_ctx);
 
   CertFreeCertificateContext(pRemoteCertContext);
-  return hash_len;
+  return (uint)ma_hash_digest_size(hash_type);
 }
 
 void ma_tls_set_connection(MYSQL *mysql __attribute__((unused)))
