@@ -399,7 +399,8 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
     goto err;
 
   if (open_cached_file(&buffpek_pointers, mysql_tmpdir, TEMP_PREFIX,
-                       DISK_CHUNK_SIZE, MYF(MY_WME)))
+                       DISK_CHUNK_SIZE,
+                       MYF(MY_WME | MY_TRACK_WITH_LIMIT)))
     goto err;
 
   param.local_sortorder=
@@ -426,6 +427,7 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
   else
   {
     /* filesort cannot handle zero-length records during merge. */
+    thd->query_plan_flags|= QPLAN_FILESORT_DISK;
     DBUG_ASSERT(param.sort_length != 0);
 
     if (sort->buffpek.str && sort->buffpek.length < maxbuffer)
@@ -451,7 +453,7 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
     /* Open cached file if it isn't open */
     if (!my_b_inited(outfile) &&
         open_cached_file(outfile, mysql_tmpdir, TEMP_PREFIX, DISK_CHUNK_SIZE,
-                         MYF(MY_WME)))
+                       MYF(MY_WME | MY_TRACK_WITH_LIMIT)))
       goto err;
     if (reinit_io_cache(outfile,WRITE_CACHE,0L,0,0))
       goto err;
@@ -548,7 +550,7 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
   else
     thd->inc_status_sort_rows(num_rows);
 
-  sort->examined_rows= param.examined_rows;
+  sort->m_examined_rows= param.examined_rows;
   sort->return_rows= num_rows;
 #ifdef SKIP_DBUG_IN_FILESORT
   DBUG_POP_EMPTY;		/* Ok to DBUG */
@@ -556,7 +558,7 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
 
   DBUG_PRINT("exit",
              ("num_rows: %lld examined_rows: %lld found_rows: %lld",
-              (longlong) sort->return_rows, (longlong) sort->examined_rows,
+              (longlong) sort->return_rows, (longlong) sort->m_examined_rows,
               (longlong) sort->found_rows));
   MYSQL_FILESORT_DONE(error, num_rows);
 
@@ -973,11 +975,6 @@ static ha_rows find_all_keys(THD *thd, Sort_param *param, SQL_SELECT *select,
     if (unlikely(thd->check_killed()))
     {
       DBUG_PRINT("info",("Sort killed by user"));
-      if (!quick_select)
-      {
-        (void) file->extra(HA_EXTRA_NO_CACHE);
-        file->ha_rnd_end();
-      }
       goto err;                               /* purecov: inspected */
     }
 
@@ -985,6 +982,7 @@ static ha_rows find_all_keys(THD *thd, Sort_param *param, SQL_SELECT *select,
     if (likely(error == 0))
     {
       param->examined_rows++;
+      thd->inc_examined_row_count_fast();
       if (select && select->cond)
       {
         /*
@@ -1081,6 +1079,11 @@ static ha_rows find_all_keys(THD *thd, Sort_param *param, SQL_SELECT *select,
   DBUG_RETURN(num_records);
 
 err:
+  if (!quick_select)
+  {
+    (void) file->extra(HA_EXTRA_NO_CACHE);
+    file->ha_rnd_end();
+  }
   sort_form->column_bitmaps_set(save_read_set, save_write_set);
   DBUG_RETURN(HA_POS_ERROR);
 } /* find_all_keys */
@@ -1119,7 +1122,7 @@ write_keys(Sort_param *param,  SORT_INFO *fs_info, uint count,
 
   if (!my_b_inited(tempfile) &&
       open_cached_file(tempfile, mysql_tmpdir, TEMP_PREFIX, DISK_CHUNK_SIZE,
-                       MYF(MY_WME)))
+                       MYF(MY_WME | MY_TRACK_WITH_LIMIT)))
     DBUG_RETURN(1);                                /* purecov: inspected */
   /* check we won't have more buffpeks than we can possibly keep in memory */
   if (my_b_tell(buffpek_pointers) + sizeof(Merge_chunk) > (ulonglong)UINT_MAX)
@@ -1561,7 +1564,7 @@ int merge_many_buff(Sort_param *param, Sort_buffer sort_buffer,
     DBUG_RETURN(0);                             /* purecov: inspected */
   if (flush_io_cache(t_file) ||
       open_cached_file(&t_file2, mysql_tmpdir, TEMP_PREFIX, DISK_CHUNK_SIZE,
-                       MYF(MY_WME)))
+                       MYF(MY_WME | MY_TRACK_WITH_LIMIT)))
     DBUG_RETURN(1);				/* purecov: inspected */
 
   from_file= t_file; to_file= &t_file2;
@@ -2201,7 +2204,8 @@ sortlength(THD *thd, Sort_keys *sort_keys, bool *allow_packing_for_sortkeys)
       set_if_smaller(sortorder->length, thd->variables.max_sort_length);
       set_if_smaller(sortorder->original_length, thd->variables.max_sort_length);
     }
-    length+=sortorder->length;
+    DBUG_ASSERT(length < UINT_MAX32 - sortorder->length);
+    length+= sortorder->length;
 
     sort_keys->increment_size_of_packable_fields(sortorder->length_bytes);
     sort_keys->increment_original_sort_length(sortorder->original_length);
