@@ -173,6 +173,7 @@ page_exists:
     page_zip_set_size(&bpage->zip, zip_size);
     bpage->zip.data = (page_zip_t*) data;
 
+    bpage->lock.init();
     bpage->init(buf_page_t::READ_FIX, page_id);
     bpage->lock.x_lock(true);
 
@@ -235,12 +236,15 @@ buf_read_page_low(
 	}
 
 	ut_ad(bpage->in_file());
-	ulonglong mariadb_timer= 0;
+	ulonglong mariadb_timer = 0;
 
 	if (sync) {
 		thd_wait_begin(nullptr, THD_WAIT_DISKIO);
-		if (mariadb_stats_active())
-		  mariadb_timer= mariadb_measure();
+		if (const ha_handler_stats *stats = mariadb_stats) {
+			if (stats->active) {
+				mariadb_timer = mariadb_measure();
+			}
+		}
 	}
 
 	DBUG_LOG("ib_buf",
@@ -253,20 +257,22 @@ buf_read_page_low(
 	auto fio = space->io(IORequest(sync
 				       ? IORequest::READ_SYNC
 				       : IORequest::READ_ASYNC),
-			     page_id.page_no() * len, len, dst, bpage);
+			     os_offset_t{page_id.page_no()} * len, len,
+			     dst, bpage);
 
 	if (UNIV_UNLIKELY(fio.err != DB_SUCCESS)) {
 		buf_pool.corrupted_evict(bpage, buf_page_t::READ_FIX);
 	} else if (sync) {
-		thd_wait_end(NULL);
+		thd_wait_end(nullptr);
 		/* The i/o was already completed in space->io() */
 		fio.err = bpage->read_complete(*fio.node);
 		space->release();
 		if (fio.err == DB_FAIL) {
 			fio.err = DB_PAGE_CORRUPTED;
 		}
-		if (mariadb_timer)
-		  mariadb_increment_pages_read_time(mariadb_timer);
+		if (mariadb_timer) {
+			mariadb_increment_pages_read_time(mariadb_timer);
+		}
 	}
 
 	return fio.err;
@@ -377,6 +383,7 @@ read_ahead:
 
   if (count)
   {
+    mariadb_increment_pages_prefetched(count);
     DBUG_PRINT("ib_buf", ("random read-ahead %zu pages from %s: %u",
 			  count, space->chain.start->name,
 			  low.page_no()));
@@ -556,11 +563,16 @@ fail:
   for (page_id_t i= low; i <= high_1; ++i)
   {
     buf_pool_t::hash_chain &chain= buf_pool.page_hash.cell_get(i.fold());
-    transactional_shared_lock_guard<page_hash_latch> g
-      {buf_pool.page_hash.lock_get(chain)};
+    page_hash_latch &hash_lock= buf_pool.page_hash.lock_get(chain);
+    /* It does not make sense to use transactional_lock_guard here,
+    because we would have many complex conditions inside the memory
+    transaction. */
+    hash_lock.lock_shared();
+
     const buf_page_t* bpage= buf_pool.page_hash.get(i, chain);
     if (!bpage)
     {
+      hash_lock.unlock_shared();
       if (i == page_id)
         goto fail;
 failed:
@@ -568,6 +580,7 @@ failed:
         continue;
       goto fail;
     }
+    const unsigned accessed= bpage->is_accessed();
     if (i == page_id)
     {
       /* Read the natural predecessor and successor page addresses from
@@ -578,6 +591,13 @@ failed:
       const byte *f= bpage->frame ? bpage->frame : bpage->zip.data;
       uint32_t prev= mach_read_from_4(my_assume_aligned<4>(f + FIL_PAGE_PREV));
       uint32_t next= mach_read_from_4(my_assume_aligned<4>(f + FIL_PAGE_NEXT));
+      hash_lock.unlock_shared();
+      /* The underlying file page of this buffer pool page could actually
+      be marked as freed, or a read of the page into the buffer pool might
+      be in progress. We may read uninitialized data here.
+      Suppress warnings of comparing uninitialized values. */
+      MEM_MAKE_DEFINED(&prev, sizeof prev);
+      MEM_MAKE_DEFINED(&next, sizeof next);
       if (prev == FIL_NULL || next == FIL_NULL)
         goto fail;
       page_id_t id= page_id;
@@ -607,8 +627,9 @@ failed:
         /* The area is not whole */
         goto fail;
     }
+    else
+      hash_lock.unlock_shared();
 
-    const unsigned accessed= bpage->is_accessed();
     if (!accessed)
       goto failed;
     /* Note that buf_page_t::is_accessed() returns the time of the
@@ -658,6 +679,7 @@ failed:
 
   if (count)
   {
+    mariadb_increment_pages_prefetched(count);
     DBUG_PRINT("ib_buf", ("random read-ahead %zu pages from %s: %u",
                           count, space->chain.start->name,
                           new_low.page_no()));
