@@ -1,15 +1,8 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2009, Google Inc.
 Copyright (c) 2014, 2023, MariaDB Corporation.
 
-Portions of this file contain modifications contributed and copyrighted by
-Google, Inc. Those modifications are gratefully acknowledged and are described
-briefly in the InnoDB documentation. The contributions by Google are
-incorporated with their permission, and subject to the conditions contained in
-the file COPYING.Google.
-
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
 Foundation; version 2 of the License.
@@ -75,9 +68,7 @@ log_t	log_sys;
 
 void log_t::set_capacity()
 {
-#ifndef SUX_LOCK_GENERIC
-	ut_ad(log_sys.latch.is_write_locked());
-#endif
+	ut_ad(log_sys.latch_have_wr());
 	/* Margin for the free space in the smallest log, before a new query
 	step which modifies the database, is started */
 
@@ -94,36 +85,53 @@ void log_t::set_capacity()
 	log_sys.max_checkpoint_age = margin;
 }
 
-/** Initialize the redo log subsystem. */
-void log_t::create()
+#ifdef HAVE_PMEM
+void log_t::create_low()
+#else
+bool log_t::create()
+#endif
 {
   ut_ad(this == &log_sys);
   ut_ad(!is_initialised());
 
-  latch.SRW_LOCK_INIT(log_latch_key);
-  init_lsn_lock();
-
   /* LSN 0 and 1 are reserved; @see buf_page_t::oldest_modification_ */
   lsn.store(FIRST_LSN, std::memory_order_relaxed);
   flushed_to_disk_lsn.store(FIRST_LSN, std::memory_order_relaxed);
+  need_checkpoint.store(true, std::memory_order_relaxed);
   write_lsn= FIRST_LSN;
 
 #ifndef HAVE_PMEM
   buf= static_cast<byte*>(ut_malloc_dontdump(buf_size, PSI_INSTRUMENT_ME));
-  TRASH_ALLOC(buf, buf_size);
+  if (!buf)
+  {
+  alloc_fail:
+    sql_print_error("InnoDB: Cannot allocate memory;"
+                    " too large innodb_log_buffer_size?");
+    return false;
+  }
   flush_buf= static_cast<byte*>(ut_malloc_dontdump(buf_size,
                                                    PSI_INSTRUMENT_ME));
+  if (!flush_buf)
+  {
+    ut_free_dodump(buf, buf_size);
+    buf= nullptr;
+    goto alloc_fail;
+  }
+
+  TRASH_ALLOC(buf, buf_size);
   TRASH_ALLOC(flush_buf, buf_size);
   checkpoint_buf= static_cast<byte*>(aligned_malloc(4096, 4096));
   memset_aligned<4096>(checkpoint_buf, 0, 4096);
+  max_buf_free= buf_size / LOG_BUF_FLUSH_RATIO - LOG_BUF_FLUSH_MARGIN;
 #else
   ut_ad(!checkpoint_buf);
   ut_ad(!buf);
   ut_ad(!flush_buf);
+  max_buf_free= 1;
 #endif
 
-  max_buf_free= buf_size / LOG_BUF_FLUSH_RATIO - LOG_BUF_FLUSH_MARGIN;
-  set_check_flush_or_checkpoint();
+  latch.SRW_LOCK_INIT(log_latch_key);
+  lsn_lock.init();
 
   last_checkpoint_lsn= FIRST_LSN;
   log_capacity= 0;
@@ -132,9 +140,12 @@ void log_t::create()
   next_checkpoint_lsn= 0;
   checkpoint_pending= false;
 
-  buf_free= 0;
+  set_buf_free(0);
 
   ut_ad(is_initialised());
+#ifndef HAVE_PMEM
+  return true;
+#endif
 }
 
 dberr_t log_file_t::close() noexcept
@@ -148,6 +159,7 @@ dberr_t log_file_t::close() noexcept
   return DB_SUCCESS;
 }
 
+__attribute__((warn_unused_result))
 dberr_t log_file_t::read(os_offset_t offset, span<byte> buf) noexcept
 {
   ut_ad(is_opened());
@@ -160,11 +172,13 @@ void log_file_t::write(os_offset_t offset, span<const byte> buf) noexcept
   ut_ad(is_opened());
   if (dberr_t err= os_file_write_func(IORequestWrite, "ib_logfile0", m_file,
                                       buf.data(), offset, buf.size()))
-    ib::fatal() << "write(\"ib_logfile0\") returned " << err;
+    ib::fatal() << "write(\"ib_logfile0\") returned " << err
+                << ". Operating system error number "
+                << IF_WIN(GetLastError(), errno) << ".";
 }
 
 #ifdef HAVE_PMEM
-# include <libpmem.h>
+# include "cache.h"
 
 /** Attempt to memory map a file.
 @param file  log file handle
@@ -200,7 +214,11 @@ static void *log_mmap(os_file_t file, os_offset_t size)
 }
 #endif
 
-void log_t::attach(log_file_t file, os_offset_t size)
+#ifdef HAVE_PMEM
+bool log_t::attach(log_file_t file, os_offset_t size)
+#else
+void log_t::attach_low(log_file_t file, os_offset_t size)
+#endif
 {
   log= file;
   ut_ad(!size || size >= START_OFFSET + SIZE_OF_FILE_CHECKPOINT);
@@ -217,19 +235,34 @@ void log_t::attach(log_file_t file, os_offset_t size)
       log.close();
       mprotect(ptr, size_t(size), PROT_READ);
       buf= static_cast<byte*>(ptr);
-#if defined __linux__ || defined _WIN32
-      set_block_size(CPU_LEVEL1_DCACHE_LINESIZE);
-#endif
+      max_buf_free= 1;
       log_maybe_unbuffered= true;
       log_buffered= false;
-      return;
+      mtr_t::finisher_update();
+      return true;
     }
   }
   buf= static_cast<byte*>(ut_malloc_dontdump(buf_size, PSI_INSTRUMENT_ME));
-  TRASH_ALLOC(buf, buf_size);
+  if (!buf)
+  {
+  alloc_fail:
+    max_buf_free= 0;
+    sql_print_error("InnoDB: Cannot allocate memory;"
+                    " too large innodb_log_buffer_size?");
+    return false;
+  }
   flush_buf= static_cast<byte*>(ut_malloc_dontdump(buf_size,
                                                    PSI_INSTRUMENT_ME));
+  if (!flush_buf)
+  {
+    ut_free_dodump(buf, buf_size);
+    buf= nullptr;
+    goto alloc_fail;
+  }
+
+  TRASH_ALLOC(buf, buf_size);
   TRASH_ALLOC(flush_buf, buf_size);
+  max_buf_free= buf_size / LOG_BUF_FLUSH_RATIO - LOG_BUF_FLUSH_MARGIN;
 #endif
 
 #if defined __linux__ || defined _WIN32
@@ -237,12 +270,17 @@ void log_t::attach(log_file_t file, os_offset_t size)
                         log_buffered
                         ? "Buffered log writes"
                         : "File system buffers for log disabled",
-                        block_size);
+                        write_size);
 #endif
 
+  mtr_t::finisher_update();
 #ifdef HAVE_PMEM
-  checkpoint_buf= static_cast<byte*>(aligned_malloc(block_size, block_size));
-  memset_aligned<64>(checkpoint_buf, 0, block_size);
+  ut_ad(ut_is_2pow(write_size));
+  ut_ad(write_size >= 512);
+  ut_ad(write_size <= 4096);
+  checkpoint_buf= static_cast<byte*>(aligned_malloc(write_size, write_size));
+  memset_aligned<512>(checkpoint_buf, 0, write_size);
+  return true;
 #endif
 }
 
@@ -274,9 +312,7 @@ void log_t::header_write(byte *buf, lsn_t lsn, bool encrypted)
 
 void log_t::create(lsn_t lsn) noexcept
 {
-#ifndef SUX_LOCK_GENERIC
-  ut_ad(latch.is_write_locked());
-#endif
+  ut_ad(latch_have_wr());
   ut_ad(!recv_no_log_write);
   ut_ad(is_latest());
   ut_ad(this == &log_sys);
@@ -293,12 +329,12 @@ void log_t::create(lsn_t lsn) noexcept
   {
     mprotect(buf, size_t(file_size), PROT_READ | PROT_WRITE);
     memset_aligned<4096>(buf, 0, 4096);
-    buf_free= START_OFFSET;
+    set_buf_free(START_OFFSET);
   }
   else
 #endif
   {
-    buf_free= 0;
+    set_buf_free(0);
     memset_aligned<4096>(flush_buf, 0, buf_size);
     memset_aligned<4096>(buf, 0, buf_size);
   }
@@ -394,7 +430,7 @@ void log_t::set_buffered(bool buffered)
                           log_buffered
                           ? "Buffered log writes"
                           : "File system buffers for log disabled",
-                          block_size);
+                          write_size);
   }
   log_resize_release();
 }
@@ -452,11 +488,12 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
     resize_lsn.store(1, std::memory_order_relaxed);
     resize_target= 0;
     resize_log.m_file=
-      os_file_create_func(path.c_str(),
-                          OS_FILE_CREATE | OS_FILE_ON_ERROR_NO_EXIT,
+      os_file_create_func(path.c_str(), OS_FILE_CREATE,
                           OS_FILE_NORMAL, OS_LOG_FILE, false, &success);
     if (success)
     {
+      ut_ad(!(size_t(file_size) & (write_size - 1)));
+      ut_ad(!(size_t(size) & (write_size - 1)));
       log_resize_release();
 
       void *ptr= nullptr, *ptr2= nullptr;
@@ -503,17 +540,14 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
         resize_target= size;
         resize_buf= static_cast<byte*>(ptr);
         resize_flush_buf= static_cast<byte*>(ptr2);
+        start_lsn= get_lsn();
+
         if (is_pmem())
-        {
           resize_log.close();
-          start_lsn= get_lsn();
-        }
         else
-        {
-          memcpy_aligned<16>(resize_buf, buf, (buf_free + 15) & ~15);
           start_lsn= first_lsn +
-            (~lsn_t{get_block_size() - 1} & (write_lsn - first_lsn));
-        }
+            (~lsn_t{write_size - 1} &
+             (lsn_t{write_size - 1} + start_lsn - first_lsn));
       }
       resize_lsn.store(start_lsn, std::memory_order_relaxed);
       status= success ? RESIZE_STARTED : RESIZE_FAILED;
@@ -568,32 +602,30 @@ void log_t::resize_abort() noexcept
 
 /** Write an aligned buffer to ib_logfile0.
 @param buf    buffer to be written
-@param len    length of data to be written
+@param length length of data to be written
 @param offset log file offset */
-static void log_write_buf(const byte *buf, size_t len, lsn_t offset)
+static void log_write_buf(const byte *buf, size_t length, lsn_t offset)
 {
   ut_ad(write_lock.is_owner());
   ut_ad(!recv_no_log_write);
-  ut_d(const size_t block_size_1= log_sys.get_block_size() - 1);
+  ut_d(const size_t block_size_1= log_sys.write_size - 1);
   ut_ad(!(offset & block_size_1));
-  ut_ad(!(len & block_size_1));
+  ut_ad(!(length & block_size_1));
   ut_ad(!(size_t(buf) & block_size_1));
-  ut_ad(len);
+  ut_ad(length);
 
-  if (UNIV_LIKELY(offset + len <= log_sys.file_size))
+  const lsn_t maximum_write_length{log_sys.file_size - offset};
+  ut_ad(maximum_write_length <= log_sys.file_size - log_sys.START_OFFSET);
+
+  if (UNIV_UNLIKELY(length > maximum_write_length))
   {
-write:
-    log_sys.log.write(offset, {buf, len});
-    return;
+    log_sys.log.write(offset, {buf, size_t(maximum_write_length)});
+    length-= size_t(maximum_write_length);
+    buf+= size_t(maximum_write_length);
+    ut_ad(log_sys.START_OFFSET + length < offset);
+    offset= log_sys.START_OFFSET;
   }
-
-  const size_t write_len= size_t(log_sys.file_size - offset);
-  log_sys.log.write(offset, {buf, write_len});
-  len-= write_len;
-  buf+= write_len;
-  ut_ad(log_sys.START_OFFSET + len < offset);
-  offset= log_sys.START_OFFSET;
-  goto write;
+  log_sys.log.write(offset, {buf, length});
 }
 
 /** Invoke commit_checkpoint_notify_ha() to notify that outstanding
@@ -768,19 +800,27 @@ inline void log_t::persist(lsn_t lsn) noexcept
 }
 #endif
 
-/** Write resize_buf to resize_log.
-@param length  the used length of resize_buf */
-ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
+ATTRIBUTE_COLD ATTRIBUTE_NOINLINE
+void log_t::resize_write_buf(const byte *b, size_t length) noexcept
 {
-  const size_t block_size_1= get_block_size() - 1;
+  const size_t block_size_1= write_size - 1;
+  ut_ad(b == resize_buf || b == resize_flush_buf);
   ut_ad(!(resize_target & block_size_1));
   ut_ad(!(length & block_size_1));
   ut_ad(length > block_size_1);
   ut_ad(length <= resize_target);
-  const lsn_t resizing{resize_in_progress()};
-  ut_ad(resizing <= write_lsn);
-  lsn_t offset= START_OFFSET +
-    ((write_lsn - resizing) & ~lsn_t{block_size_1}) %
+
+  int64_t d= int64_t(write_lsn - resize_in_progress());
+  if (UNIV_UNLIKELY(d <= 0))
+  {
+    d&= ~int64_t(block_size_1);
+    if (int64_t(d + length) <= 0)
+      return;
+    length+= d;
+    b-= d;
+    d= 0;
+  }
+  lsn_t offset= START_OFFSET + (lsn_t(d) & ~lsn_t{block_size_1}) %
     (resize_target - START_OFFSET);
 
   if (UNIV_UNLIKELY(offset + length > resize_target))
@@ -792,7 +832,7 @@ ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
   }
 
   ut_a(os_file_write_func(IORequestWrite, "ib_logfile101", resize_log.m_file,
-                          resize_flush_buf, offset, length) == DB_SUCCESS);
+                          b, offset, length) == DB_SUCCESS);
 }
 
 /** Write buf to ib_logfile0.
@@ -800,11 +840,9 @@ ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
 @return the current log sequence number */
 template<bool release_latch> inline lsn_t log_t::write_buf() noexcept
 {
-#ifndef SUX_LOCK_GENERIC
-  ut_ad(latch.is_write_locked());
-#endif
-  ut_ad(!srv_read_only_mode);
+  ut_ad(latch_have_wr());
   ut_ad(!is_pmem());
+  ut_ad(!srv_read_only_mode);
 
   const lsn_t lsn{get_lsn(std::memory_order_relaxed)};
 
@@ -816,66 +854,92 @@ template<bool release_latch> inline lsn_t log_t::write_buf() noexcept
   }
   else
   {
+    ut_ad(write_lock.is_owner());
     ut_ad(!recv_no_log_write);
     write_lock.set_pending(lsn);
     ut_ad(write_lsn >= get_flushed_lsn());
-    const size_t block_size_1{get_block_size() - 1};
-    lsn_t offset{calc_lsn_offset(write_lsn) & ~lsn_t{block_size_1}};
+    const size_t write_size_1{write_size - 1};
+    ut_ad(ut_is_2pow(write_size));
+    size_t length{buf_free.load(std::memory_order_relaxed)};
+    lsn_t offset{calc_lsn_offset(write_lsn)};
+    ut_ad(length >= (offset & write_size_1));
+    ut_ad(write_size_1 >= 511);
 
-    DBUG_PRINT("ib_log", ("write " LSN_PF " to " LSN_PF " at " LSN_PF,
-                          write_lsn, lsn, offset));
-    const byte *write_buf{buf};
-    size_t length{buf_free};
-    ut_ad(length >= (calc_lsn_offset(write_lsn) & block_size_1));
-    const size_t new_buf_free{length & block_size_1};
-    buf_free= new_buf_free;
-    ut_ad(new_buf_free == ((lsn - first_lsn) & block_size_1));
+    const byte *const write_buf{buf};
+    const byte *const re_write_buf{resize_buf};
+    offset&= ~lsn_t{write_size_1};
 
-    if (new_buf_free)
+    if (length <= write_size_1)
     {
+      ut_ad(!((length ^ (size_t(lsn) - size_t(first_lsn))) & write_size_1));
+      /* Keep filling the same buffer until we have more than one block. */
 #if 0 /* TODO: Pad the last log block with dummy records. */
-      buf_free= log_pad(lsn, get_block_size() - new_buf_free,
-                        buf + new_buf_free, flush_buf);
+      buf_free= log_pad(lsn, (write_size_1 + 1) - length,
+                        buf + length, flush_buf);
       ... /* TODO: Update the LSN and adjust other code. */
 #else
-      /* The rest of the block will be written as garbage.
-      (We want to avoid memset() while holding mutex.)
-      This block will be overwritten later, once records beyond
-      the current LSN are generated. */
 # ifdef HAVE_valgrind
-      MEM_MAKE_DEFINED(buf + length, get_block_size() - new_buf_free);
-      if (UNIV_LIKELY_NULL(resize_flush_buf))
-        MEM_MAKE_DEFINED(resize_buf + length, get_block_size() - new_buf_free);
+      MEM_MAKE_DEFINED(buf + length, (write_size_1 + 1) - length);
+      if (UNIV_LIKELY_NULL(re_write_buf))
+        MEM_MAKE_DEFINED(re_write_buf + length, (write_size_1 + 1) - length);
 # endif
       buf[length]= 0; /* allow recovery to catch EOF faster */
-      length&= ~block_size_1;
-      memcpy_aligned<16>(flush_buf, buf + length, (new_buf_free + 15) & ~15);
-      if (UNIV_LIKELY_NULL(resize_flush_buf))
-        memcpy_aligned<16>(resize_flush_buf, resize_buf + length,
-                           (new_buf_free + 15) & ~15);
-      length+= get_block_size();
 #endif
+      length= write_size_1 + 1;
+    }
+    else
+    {
+      const size_t new_buf_free{length & write_size_1};
+      ut_ad(new_buf_free == ((lsn - first_lsn) & write_size_1));
+      buf_free.store(new_buf_free, std::memory_order_relaxed);
+
+      if (new_buf_free)
+      {
+        /* The rest of the block will be written as garbage.
+        (We want to avoid memset() while holding exclusive log_sys.latch)
+        This block will be overwritten later, once records beyond
+        the current LSN are generated. */
+#ifdef HAVE_valgrind
+          MEM_MAKE_DEFINED(buf + length, (write_size_1 + 1) - new_buf_free);
+        if (UNIV_LIKELY_NULL(re_write_buf))
+          MEM_MAKE_DEFINED(re_write_buf + length, (write_size_1 + 1) -
+                           new_buf_free);
+#endif
+        buf[length]= 0; /* allow recovery to catch EOF faster */
+        length&= ~write_size_1;
+        memcpy_aligned<16>(flush_buf, buf + length, (new_buf_free + 15) & ~15);
+        if (UNIV_LIKELY_NULL(re_write_buf))
+          memcpy_aligned<16>(resize_flush_buf, re_write_buf + length,
+                             (new_buf_free + 15) & ~15);
+        length+= write_size_1 + 1;
+      }
+
+      std::swap(buf, flush_buf);
+      std::swap(resize_buf, resize_flush_buf);
     }
 
-    std::swap(buf, flush_buf);
-    std::swap(resize_buf, resize_flush_buf);
     write_to_log++;
     if (release_latch)
       latch.wr_unlock();
 
+    DBUG_PRINT("ib_log", ("write " LSN_PF " to " LSN_PF " at " LSN_PF,
+                          write_lsn, lsn, offset));
+
+    /* Do the write to the log file */
+    log_write_buf(write_buf, length, offset);
+
+    if (UNIV_LIKELY_NULL(re_write_buf))
+      resize_write_buf(re_write_buf, length);
+    write_lsn= lsn;
+
     if (UNIV_UNLIKELY(srv_shutdown_state > SRV_SHUTDOWN_INITIATED))
     {
       service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
                                      "InnoDB log write: " LSN_PF, write_lsn);
     }
-
-    /* Do the write to the log file */
-    log_write_buf(write_buf, length, offset);
-    if (UNIV_LIKELY_NULL(resize_buf))
-      resize_write_buf(length);
-    write_lsn= lsn;
   }
 
+  set_check_for_checkpoint(false);
   return lsn;
 }
 
@@ -917,8 +981,9 @@ wait and check if an already running write is covering the request.
 void log_write_up_to(lsn_t lsn, bool durable,
                      const completion_callback *callback)
 {
-  ut_ad(!srv_read_only_mode);
+  ut_ad(!srv_read_only_mode || log_sys.buf_free_ok());
   ut_ad(lsn != LSN_MAX);
+  ut_ad(lsn != 0);
   ut_ad(lsn <= log_sys.get_lsn());
 
 #ifdef HAVE_PMEM
@@ -967,7 +1032,6 @@ repeat:
 @param durable  whether to wait for a durable write to complete */
 void log_buffer_flush_to_disk(bool durable)
 {
-  ut_ad(!srv_read_only_mode);
   log_write_up_to(log_sys.get_lsn(std::memory_order_acquire), durable);
 }
 
@@ -999,16 +1063,6 @@ ATTRIBUTE_COLD void log_write_and_flush()
 #endif
 }
 
-/********************************************************************
-
-Tries to establish a big enough margin of free space in the log buffer, such
-that a new log entry can be catenated without an immediate need for a flush. */
-ATTRIBUTE_COLD static void log_flush_margin()
-{
-  if (log_sys.buf_free > log_sys.max_buf_free)
-    log_buffer_flush_to_disk(false);
-}
-
 /****************************************************************//**
 Tries to establish a big enough margin of free space in the log, such
 that a new log entry can be catenated without an immediate need for a
@@ -1016,12 +1070,12 @@ checkpoint. NOTE: this function may only be called if the calling thread
 owns no synchronization objects! */
 ATTRIBUTE_COLD static void log_checkpoint_margin()
 {
-  while (log_sys.check_flush_or_checkpoint())
+  while (log_sys.check_for_checkpoint())
   {
     log_sys.latch.rd_lock(SRW_LOCK_CALL);
     ut_ad(!recv_no_log_write);
 
-    if (!log_sys.check_flush_or_checkpoint())
+    if (!log_sys.check_for_checkpoint())
     {
 func_exit:
       log_sys.latch.rd_unlock();
@@ -1037,7 +1091,7 @@ func_exit:
 #ifndef DBUG_OFF
     skip_checkpoint:
 #endif
-      log_sys.set_check_flush_or_checkpoint(false);
+      log_sys.set_check_for_checkpoint(false);
       goto func_exit;
     }
 
@@ -1051,30 +1105,17 @@ func_exit:
   }
 }
 
-/**
-Checks that there is enough free space in the log to start a new query step.
-Flushes the log buffer or makes a new checkpoint if necessary. NOTE: this
-function may only be called if the calling thread owns no synchronization
-objects! */
-ATTRIBUTE_COLD void log_check_margins()
-{
-  do
-  {
-    log_flush_margin();
-    log_checkpoint_margin();
-    ut_ad(!recv_no_log_write);
-  }
-  while (log_sys.check_flush_or_checkpoint());
-}
-
 /** Wait for a log checkpoint if needed.
 NOTE that this function may only be called while not holding
 any synchronization objects except dict_sys.latch. */
 void log_free_check()
 {
-  ut_ad(!lock_sys.is_writer());
-  if (log_sys.check_flush_or_checkpoint())
-    log_check_margins();
+  ut_ad(!lock_sys.is_holder());
+  if (log_sys.check_for_checkpoint())
+  {
+    ut_ad(!recv_no_log_write);
+    log_checkpoint_margin();
+  }
 }
 
 extern void buf_resize_shutdown();
@@ -1140,9 +1181,7 @@ loop:
 	}
 
 	/* We need these threads to stop early in shutdown. */
-	const char* thread_name = srv_fast_shutdown != 2
-		&& trx_rollback_is_active
-		? "rollback of recovered transactions" : nullptr;
+	const char* thread_name= nullptr;
 
 	if (thread_name) {
 		ut_ad(!srv_read_only_mode);
@@ -1174,6 +1213,7 @@ wait_suspend_loop:
 	}
 
 	buf_load_dump_end();
+	rollback_all_recovered_task.wait();
 
 	if (!buf_pool.is_initialised()) {
 		ut_ad(!srv_was_started);
@@ -1285,6 +1325,7 @@ log_print(
 void log_t::close()
 {
   ut_ad(this == &log_sys);
+  ut_ad(!(buf_free & buf_free_LOCK));
   if (!is_initialised()) return;
   close_file();
 
@@ -1302,7 +1343,7 @@ void log_t::close()
 #endif
 
   latch.destroy();
-  destroy_lsn_lock();
+  lsn_lock.destroy();
 
   recv_sys.close();
 
